"""spyndex - Awesome Spectral Indices in Python"""

__version__ = "0.10.0"
__author__ = "David Montero Loaiza <dml.mont@gmail.com>"
__all__ = []

from . import datasets, plot
from .axioms import bands, constants, indices
from .spyndex import *
